## argmax of f prediction
#  use veterbi algorithm on tree
function argmax_f_opt_var(T::Vector{Int}, PS_node::Matrix{Float64}, PS_edge::Array{Float64,3})

    nc, nn = size(PS_node)

    Z = zeros(Int, nc, nn)          # store argmax
    F = zeros(nc, nn)               # store max

    # find max
    @inbounds @fastmath for i = nn:-1:1
        ch_ids = findall(T[i+1:end] .== i)
        if length(ch_ids) == 0          # leaf node
            for k = 1:nc                # each class
                imax = 0
                vmax = -Inf
                @simd for l = 1:nc
                    candidate = PS_node[l,i] + PS_edge[k,l,i] + PS_node[k,T[i]]
                    if candidate > vmax
                        imax = l
                        vmax = candidate
                    end
                end
                F[k,i] = vmax
                Z[k,i] = imax
            end
        else                            # has childeren
            for k = 1:nc                # each class
                imax = 0
                vmax = -Inf
                for l = 1:nc
                    candidate = 0.0
                    @simd for j in ch_ids         
                        candidate += F[l, i+j] - PS_node[l,i]
                    end
                    if i == 1
                        candidate += PS_node[l,i] + PS_edge[k,l,i] 
                    else
                        candidate += PS_node[l,i] + PS_edge[k,l,i] + PS_node[k,T[i]]
                    end

                    if candidate > vmax
                        imax = l
                        vmax = candidate
                    end
                end
                F[k,i] = vmax
                Z[k,i] = imax
            end
        end
    end

    # backtrack
    val = F[1,1]
    sol = zeros(Int, nn)
    sol[1] = Z[1,1]
    @inbounds for i = 2:nn
        sol[i] = Z[sol[T[i]],i]
    end
    
    return sol, val
end


function test_argmax_f_opt_var()
    @time begin
        for j = 1:1000
            nc = 10
            nn = 100
            
            T = zeros(Int, nn)
            T[1] = 0
            T[2] = 1
            for i = 3:nn
                T[i] = rand(1:i-1)
            end
            
            PS_node = rand(nc,nn)
            PS_edge = rand(nc,nc,nn)
            PS_edge[:,:,1] = zeros(nc,nc)

            sol1, v1 = argmax_f_opt_var(T, PS_node, PS_edge)
        end
    end
end


