##### 1
# Graphical Model cases
# AGM vs CRF vs SVM
# argmax f

##### 2 
# Programming Languages
# difference performance between compiled & interpreted
# test_sum 

##### Linear Algebra 
# BLAS & LAPACK


##### 4
# Julia intro
# Julia benchmark site

# syntax basic
a = 1
a = 1;
b = 1.0
c = true

γ = 1 // 2
α = 2b

v = [1, 2, 3]
v = [1.0, 1.2, 0.2]

A = [1 2 3; 4 5 6; 7 8 9]

v = [1, 2]
w = [3, 4]
B = [v w]
C = [v; w]
D = [v, w]

typeof(a)
typeof(b)
typeof(c)
1/2
typeof(1 / 2)
1//2
typeof(1 // 2)

supertype(Float64)
supertype(supertype(Float64))
subtypes(Real)
subtypes(Number)

Float64 <: Number
Int64 <: Real
typeof(pi)
typeof(pi) <: Rational

v = zeros(5)
X = rand(5)
X = rand(5,5)
Y = ones(2,2,2)

Z = [2i for i=1:3]
Z = [i*j for i=1:3, j=1:2]

X[1,1]
X[:,1]
X[:,1:2]
X[4:end,1:2]

X = ones(3,3)
Y = rand(3,3)
v = [1,2,3]
w = zeros(3)

X * Y
X .* Y
v * w
v .* w
v' * w
X * v

X + Y
X + v
X .+ v
X .+ v'

s = "This is string"
t = " This one too"
s * t
s + t
s = 'This is string'
ch = 'a'
println("Var a is $a, while b plus 1 is $(b+1)")

## used to be more similar to matlab
## v 1.0 -> create its version
## Make base Julia small, 
## using LinearAlgebra
det(X)
std(v)

using LinearAlgebra
det(X)
using Statistics
std(v)


##### 5
# Function and type stability
square2(x) = 2x * x
square2(2)
square2(1.5)
square2([1 2; 3 4])

@code_typed square2(2)
@code_typed square2(1.5)
@code_typed square2([1 2; 3 4])

@code_native square2(2)
@code_native square2(1.5)
@code_native square2([1 2; 3 4])


# stability
function sumrand(n)
    res = 1.0
    for i = 1:n
        res += rand()
    end
    return res
end

function sumrand2(n)
    res = 1
    for i = 1:n
        res += rand()
    end
    return res
end

function sumrand3(n)
    res = 1//1
    for i = 1:n
        res += rand()
    end
    return res
end

@time [sumrand(100_000) for i in 1:1000];
@time [sumrand2(100_000) for i in 1:1000];
@time [sumrand3(100_000) for i in 1:1000];

@code_warntype sumrand(100_000)
@code_warntype sumrand2(100_000)
@code_warntype sumrand3(100_000)


function sumpi(n)
    res = 1.0
    for i = 1:n
        res += pi
    end
    return res
end

function sumpi2(n)
    res = 1
    for i = 1:n
        res += pi
    end
    return res
end

@code_native sumpi(10)
@code_native sumpi2(10)

## REMINDER: ALWAYS WRITE FUNCTION. 
## GLOBAL is slow

##### 6
# multiple dispatch

function print_var(var::Int)
    println("This is an integer with value = $var")
end

function print_var(var::Real)
    println("This is a real variable with value = $var")
end

function print_var(var::String)
    println("This is a string with value = $var")
end

function print_var(var)
    println("This is a variable with value = $var")
end

print_var(1)
print_var(1.0)
print_var("aaa")
print_var(1 + 1im)

@which print_var(1.0)
@which print_var(1 + 1im)

print_var
methods(print_var)


@which 1 + 1
@which 2.0 + 1
@which rand(2) + rand(2)
+
methods(+)


##### 6
# import python
X = rand(30, 5)
y = rand(1:2, 30)

using PyCall

# python's code
# import sklearn.svm as sv
# clf = sv.SVC(gamma = "auto", C = 1.0)
# clf.fit(X, y)
# y_pred = clf.predict(X_test)

@pyimport sklearn.svm as sv
clf = sv.SVC(gamma = "auto", C = 1.0)
clf[:fit](X, y)
y_pred = clf[:predict](X)
error = sum(y .== y_pred) / length(y)


##### 7
# dot broadcast

# broadcast
v = collect(1:10)
X = ones(10,10)

X * v
X .* v
X .* v'

# broadcast function
fun(x) = sqrt(x) + log(x) + 1
fun(1)
fun(v)
fun.(v)
fun.(X)

# builtin function
exp(v)
exp.(v)
sin(v)
sin.(v)
v + 1
v .+ 1
v + I
X - 2
X .- 2

# dot fusion
x = rand(50_000_000)

function calc(x)
    y = 2 * x.^2 + 6 * x.^3 - sqrt.(x)
    return y
end
@time calc(x);

function calc2(x)
    n = length(x)
    y = zeros(n)
    for i = 1:n
        y[i] = 2 * x[i]^2 + 6 * x[i]^3 - sqrt(x[i])
    end
    return y
end
@time calc2(x);

calc3(x) = 2 .* x.^2 .+ 6 .* x.^3 .- sqrt.(x)
@time calc3(x);

calc4(x) = 2x^2 + 6x^3 - sqrt(x)
@time calc4(x);
@time calc4.(x);

# matlab
# x = rand(50000000,1);
# tic
# y = 2 * x.^2 + 6 * x.^3 - sqrt(x);
# toc
# see br.m


##### 8
# back to optimizing argmax


### other
# broadcast, function, dot fusion, parallel, IDE, plots, REPL tricks (unicode plot)